#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <libusb-1.0/libusb.h>
#include "usb.h"
#include "mlog.h"
#include "../imports/buffer.h"


static uint32_t EP_control    = 0x00; // host -> controller : HCI commands
static uint32_t EP_intr       = 0x81; // controller -> host : HCI events
static uint32_t EP_bulk_read  = 0x82; // controller -> host : ACL
static uint32_t EP_bulk_write = 0x02; // host -> controller : ACL

static libusb_device_handle 
               *device_handle;

static int      ulog           (int level, const char *msg,int eno);
static int      is_bt_device   (libusb_device *device);
static int      get_device_info(libusb_device *device,
                                uint16_t *idVendor, uint16_t *idProduct,
                                char **serial, char **description);
static char    *get_string     (libusb_device_handle *han, uint32_t num);
static char    *strcom         (const char *first, ...);
static int      try_open       (libusb_device *device,
                                uint32_t vendor, uint32_t product,
                                char *serial);

int usb_init()
{
  int r;
  r= libusb_init(NULL);
  if (!r) return 0;
  ulog(1,"Can't initialize libusb",r);
  return 1;
}

int usb_get_device_list(usb_device_t **Rdevices, int *Ndevices)
{
  libusb_device **list;
  buffer_t *devices;
  ssize_t nD;
  int i,nB;

  nD= libusb_get_device_list(NULL, &list);
  if (nD<0)
  {
    ulog(1,"can't get device list",nD);
    return 1;
  }

  devices= buffer_new(128);
  nB= 0;

  for(i=0;i<nD;i++)
    if (is_bt_device(list[i]))
    {
      usb_device_t D;
      if (get_device_info(list[i], &D.vendor, &D.product,
                                   &D.serial, &D.descr)) continue;
      nB++;
      buffer_write(devices, &D, sizeof(D));
    }
  buffer_free(devices, (void**) Rdevices, NULL);
  *Ndevices= nB;
  libusb_free_device_list(list,1);
  if (nB==0) ulog(2, "no bluetooth devices",0);
  return 0;
}


static int ulog(int level, const char *msg,int eno)
{
  mlog(level, "USB:%s - %s [%d]\n", msg, libusb_strerror(eno),eno);
  return 1;
}


static int is_bt_device(libusb_device *dev)
{
  struct libusb_device_descriptor devdes;
  memset(&devdes, 0, sizeof(devdes));
  libusb_get_device_descriptor(dev, &devdes);
      /* These values are from BTSPEC-VOL4-CHAP3 */
  return devdes.bDeviceClass==    0xE0
      && devdes.bDeviceSubClass== 0x01
      && devdes.bDeviceProtocol== 0x01  ;
}

static int get_device_info(libusb_device *dev,
                            uint16_t *idVendor, uint16_t *idProduct,
                            char **serial, char **description)
{
  struct libusb_device_descriptor devdes;
  libusb_device_handle *h;
  char *P, *M;
  int r;

  libusb_get_device_descriptor(dev, &devdes);
  *idVendor= devdes.idVendor;
  *idProduct= devdes.idProduct;

  *serial= *description= NULL;

  r=libusb_open(dev, &h);
  if (r)
     return ulog(2,"can't open device for info",r);

  *serial= get_string(h, devdes.iSerialNumber);
  M= get_string(h,devdes.iManufacturer);
  P= get_string(h,devdes.iProduct);

  switch((M?2:0) | (P?1:0))
  {
  case 0: *description= strdup("Unnamed Bluetooth Device"); break;
  case 1: *description= P; break;
  case 2: *description= M; break;
  case 3: *description= strcom(M,"/",P,NULL); free(P); free(M); break;
  }

  libusb_close(h);
  return 0;
}

static char *get_string(libusb_device_handle *han, uint32_t num)
{
  char buf[1024];
  int n;
  if (!num) return NULL;
  n= libusb_get_string_descriptor_ascii(han, num, (unsigned char*)buf,
                                        sizeof(buf)-1);
  if (n<0) return NULL;
  if (n>sizeof(buf)-1) n= sizeof(buf)-1;
  buf[n]= 0;
  return strdup(buf);
}

static char *strcom(const char *first, ...)
{
  va_list ap;
  const char *X;
  size_t len,L;
  char *buf;

  buf= NULL;
again:
  X= first;
  len= 0;
  va_start(ap,first);
  while(X)
  {
    L= strlen(X);
    if (buf) memcpy(buf+len,X, L);
    len+= L;
    X= va_arg(ap,const char*);
  }
  va_end(ap);
  if (buf) { buf[len]=0; return buf; }
  buf= malloc(len+1);
  goto again;
}


/*************************************************************************
  Opening a USB device is quite trivial, but I'm not sure about which
  configuration to use for other devices. For now, I'm just going to
  use the following:
    
    - use configuration 0
    - detach kernel driver
    - claim interfaces 0 and 1
    - set alternate 0 for interface 1

  Later on, I might add some 'driver' string for each device which specifies
  what to do when opening the device. For instance, we can have:

    0:0-1/0-2

  which would set configuration 0, claim interfaces 0,1,2 and set alternate
  0 for interface 1.

  In any case, let's do it without the driver stuff first and be done
  with it. 
*************************************************************************/

static int try_open
  (libusb_device *device, uint32_t vendor, uint32_t product, char *serial)
{
  int r;
  char* S;
  struct libusb_device_descriptor devdes;
  libusb_get_device_descriptor(device, &devdes);
  
  if (devdes.idVendor!=vendor) return 1;
  if (devdes.idProduct!=product) return 1;
  if (devdes.iSerialNumber==0 && serial!=NULL) return 1;
  if (devdes.iSerialNumber!=0 && serial==NULL) return 1;

  r=libusb_open(device, &device_handle); 
  if (r)
     return ulog(1,"can't open device", r);

  if (!serial) goto serial_done;
 
  S= get_string(device_handle, devdes.iSerialNumber);
  if (!S || strcmp(serial,S)) goto serial_fail;
  free(S);
serial_done:
  libusb_set_auto_detach_kernel_driver(device_handle, 1);
  if ((r= libusb_set_configuration(device_handle, 0))) goto config_fail;
  if ((r= libusb_claim_interface(device_handle, 0)))   goto config_fail;

  /* the following two are not really necessary, I just don't want
     someone else messing with the voice channels */
  libusb_claim_interface(device_handle, 1);
  libusb_set_interface_alt_setting(device_handle, 1, 0);
  
  return 0;

serial_fail:
  if (S) free(S);
  libusb_close(device_handle);
  return 1;

config_fail:
  libusb_close(device_handle);
  return ulog(1,"device config failed",r);
}



int usb_open(uint32_t vendor,uint32_t product, char *serial) 
{
  ssize_t nD;
  libusb_device **list;
  int i;

  nD= libusb_get_device_list(NULL, &list);
  if (nD<0)
     return ulog(1,"can't get device list",nD);
  if (nD)
  {
    for(i=0;i<nD;i++)
      if (try_open(list[i], vendor,product, serial)==0)
      {
        libusb_free_device_list(list,1);
        return 0;
      }
    libusb_free_device_list(list,1);
  }
  return ulog(1,"can't open device",LIBUSB_ERROR_NO_DEVICE);
}
