#include <stdarg.h>
#include <stdint.h>
#include <string.h>

/* 
this function creates a byte array from the given arguments. Each argument
is specified in a printf-like way, but it's more flexible. You need to first
give a format string which describes the arguments following it. A format
string can specify one or more arguments. The end of the whole argument
list is indicated by a NULL format string.

The meanings of format characters are:

   w -> 2 bytes, little endian
   W -> 2 bytes, big endian
   b -> byte
   d,D-> same as wW, 4 bytes
   lw -> length(word) of next group, 2 bytes, little endian
   lW -> length(word) of next group, 2 bytes, big endian
   ld -> same as lw, 4 bytes, little endian
   lD -> same as lW, 4 bytes, big endian
   lb -> same as l2, 1 bytes
    p -> array of bytes, the length comes before it
    a -> a va_list, for nested stuff.
    [ -> group begin
    ] -> group end

You can give multiple format strings seperated by arguments. For example,
the following is valid:

  bpack(buffer, 255, "bw", 3, 2016, "b", 5, NULL);

The last format string must be NULL as shown above. For pointer arguments,
you need to give the length of the byte buffer to be copied before the 
actual pointer value. For instance:

  bpack(buffer, 255, "p", 6, addr);

is good. The function returns the number of bytes copied into 'buffer'.
If this is greater than blen, then nothing is copied but it is still returned.
You must check for this condition yourself. */


int vbpack(uint8_t *dst, int blen, va_list oap)
{
   int off;
   uint8_t *parg;
   va_list aarg, ap;
   uint32_t iarg;
   char *fmt;

   int nparen;
   int parens[30];
   int parname;
   int parsz[30];
   int P;


   //printf("Entered VBAR (%d)\n", blen);
   //printf("sz(o)= %d sz(a)= %d\n", sizeof(oap), sizeof(ap));

   va_copy(ap,oap);
   off= 0;

   fmt= va_arg(ap, char*);

   // we first figure out the sizes of the parenthesised expressions.
   // everytime we find a left paren character, we increase the parname
   // which is the reference to this parenthesised expression. we
   // initialize parsz to the current offset. we also push the new
   // parenthesised expression name so that we can later match the
   // closing right-paren character.

   nparen= 0;
   parname= -1;

   while(fmt)
   {
     switch(*(fmt++))
     {
      case 'b': iarg= va_arg(ap,uint32_t); off++; break;
      case 'W': case 'w': iarg= va_arg(ap,uint32_t); off+=2; break;
      case 'd': case 'D': iarg= va_arg(ap,uint32_t); off+=4; break;
      case 'l': if      (fmt[0]=='w' || fmt[0]=='W') off+= 2; 
                else if (fmt[0]=='d' || fmt[0]=='D') off+= 4;
                else if (fmt[0]=='b')                off+= 1;
                fmt++;
                break;
      case 'p': iarg= va_arg(ap,uint32_t);
                parg= va_arg(ap, uint8_t*); 
                off+= iarg;
                break;
      case 'a': // va_list HACK
                parg= va_arg(ap, uint8_t*);
                memcpy(&aarg, parg, sizeof(aarg));
                off+= vbpack(dst+off, 0, aarg);
                break;
      case '[': parname++; parens[nparen++]= parname;
                parsz[parname]= off; break;
      case ']': P= parens[--nparen]; //  FIXME: check for underflow
                parsz[P]= off - parsz[P];
                break; 
     }
     if (!fmt[0]) fmt= va_arg(ap, char*);
   } 
   va_end(ap);

   // if (off>blen) printf("Exiting VBAR - S\n");
   if (off>blen) return off;

   va_copy(ap,oap);
   off= 0;
   parname= -1;

   fmt= va_arg(ap, char*);
 
   while(fmt)
   {
     switch(*fmt)
     {
      case 'b': 
      case 'W': 
      case 'w':
      case 'D':
      case 'd': iarg= va_arg(ap,uint32_t); break;
      case 'l': iarg= parsz[parname+1]; fmt++; break;
      case 'p': iarg= va_arg(ap,uint32_t); break;
      case '[': parname++; break;
     }

     switch(*(fmt++))
     {
      case 'b': dst[off]= iarg&0xff; off++; break;
      case 'w':
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          break;
      case 'd':
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          dst[off]= iarg & 0xff; off++; iarg>>= 8;
          break;
      case 'W':
          iarg <<= 16;
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          break;
      case 'D':
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          dst[off]= (iarg >> 24) & 0xff; off++; iarg<<= 8;
          break;
      case 'p':
          parg= va_arg(ap, uint8_t*);
          memcpy(dst+off, parg, iarg);
          off+= iarg;
          break;
      case 'a': // va_list HACK
          parg= va_arg(ap, uint8_t*);
          memcpy(&aarg, parg, sizeof(aarg));
      //    aarg= va_arg(ap, va_list);
          off+= vbpack(dst+off, blen-off, aarg);
          break;
     }
     if (!fmt[0]) fmt= va_arg(ap, char*);
   } 
   va_end(ap);
   //printf("Exiting VBAR - C\n");
   return off;
}

int bpack(uint8_t *dst, int blen, ...)
{
   int R;
   va_list ap;
   va_start(ap, blen);
   R= vbpack(dst, blen, ap);
   va_end(ap);
   return R;
}

/*
In the above code, there are two places marked as 'va_list HACK'.

These lines are specific to gcc on amd64 because of how gcc implements the
va_* macros. The trouble is, we can't simply get a va_list value from another
va_list. when you
  - declare a va_list as a local variable, it's declared as a struct.
  - try to get a va_list from another va_list, it tries to get the whole
    struct
  - pass a va_list as an argument to another function its *address*
    is passed. so, if you uncomment the second printf in vbpack (the one
    with sizeof), you will see that size of oap is 8 while it's 24 for ap.
    Although they are declared with the same name, the actual type of
    - oap is a pointer
    - ap is a struct
    
So, I simply get the pointer from va_list instead of the whole struct (which
was not passed in) and then do the copy myself using memcpy. On other 
platforms where there is no need for register spills (ie. all arguments are
already on the stack), this will have to change. This is because the va_list
type is probably a simple pointer. In that case, it will be passed and
declared with the same size and type.
*/

/* 
We have another function, mbpack. This one allocates enough bytes to
store the given contents and then goes on with formatting. The returned
value is the size of the buffer. The actual buffer is stored into the
first pointer arg. */

int mbpack(uint8_t **dst, ...)
{
   int R;
   uint8_t *buffer;
   va_list ap;

   va_start(ap, dst);
   R= vbpack(NULL, 0, ap);
   va_end(ap);

   buffer= malloc(R);
   va_start(ap, dst);
   vbpack(buffer, R, ap);
   va_end(ap);

   return R;
}
