#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

typedef struct type {
  struct type *next;
  char *name;
  char *cname;
  char *destroy;
} type_t;

typedef struct arg
{
  struct arg *next;
  char *name;
  type_t *type;
} arg_t;


typedef struct msg
{
  struct msg *next;
  char *name;
  arg_t *args;
} msg_t;

type_t *types;
msg_t *msgs;


msg_t **msg_tail= &msgs;
type_t **type_tail= &types;

char *header= "msg.h";
char *source= "msg.c";
char *pfx= "msg";

void strip_spaces(char *buf)
{
  int L;
  while(buf[0] && !isgraph(buf[0]))
    memmove(buf, buf+1, strlen(buf+1)+1);
  L= strlen(buf);
  while(L>0)
  {
    if (isgraph(buf[L-1])) break;
    buf[L-1]= 0;
    L--;
  }
}

static void new_type(char *name, char *cname, char *destroy)
{
       type_t *T= malloc(sizeof(*T));
       T->name= strdup(name);
       T->cname= strdup(cname);
       T->destroy= strdup(destroy);
       T->next= 0;
       *type_tail= T;
       type_tail= &T->next;
}

msg_t *new_msg(char *name)
{
    msg_t *M;
    M= malloc(sizeof(*M));
    M->name= strdup(name);
    M->next= 0; 
    M->args= 0;
    *msg_tail= M;
    msg_tail= &M->next;
    return M;
}

char *get_word(char *begin, char *end)
{
  char *e;
  while(begin<end)
  {
    if (isgraph(begin[0])) break;
    begin++;
  }
  if (begin>=end) return 0;
  for(e=begin+1;e<end && e[0];e++) 
      if (!isgraph(e[0])) break;
  e[0]= 0;
  return begin;
}

void parse_file(char *fn)
{
  FILE *f= fopen(fn,"r");
  msg_t *msg=0;
  type_t *type=0;
  arg_t **arg_tail= 0;
  int line= 0;
  char buf[2000];
  int i;

  if (!f) exit(fprintf(stderr,"can't open %s: %s\n", fn, strerror(errno)));
  while(1)
  {
    char *cmd,*arg[3];
    if (!fgets(buf, sizeof(buf)-1, f)) break;
    line++;
    for(i=0;buf[i];i++) if (buf[i]=='#') { buf[i]= 0; break; }
    strip_spaces(buf);
    if (!buf[0]) continue;
    if (buf[0]!='@')
        exit(fprintf(stderr,"%s:%d: syntax error, @ expected\n",fn, line));
    char *end= buf+strlen(buf);
    char *begin=buf;
    cmd= get_word(buf,end);
    if (!strcmp(cmd,"@type"))
    {
       begin= cmd+strlen(cmd)+1;
       for(i=0;i<2;i++) {
         arg[i]= get_word(begin, end);
         if (!arg[i])
            exit(fprintf(stderr,"%s:%d: syntax error, @type command needs "
                 " three arguments\n",fn, line));
         begin= arg[i]+ strlen(arg[i])+1;
       }
       arg[2]= begin;
       if (arg[2]>=end)
            exit(fprintf(stderr,"%s:%d: syntax error, @type command needs "
                 " three arguments.\n",fn, line));
       new_type(arg[0], arg[2], arg[1]);
       arg_tail= 0;
       continue;
    } else if (!strcmp(cmd,"@msg")) {
       arg[0]= get_word(cmd+strlen(cmd)+1, end);
       if (!arg[0])
           exit(fprintf(stderr,"%s:%d: syntax error, @msg command needs "
                    "an argument.\n", fn, line));
       msg_t *M= new_msg(arg[0]);
       arg_tail= &M->args;
       continue;
    } else if (!strcmp(cmd,"@source")) {
       source= strdup(cmd+strlen(cmd)+1);
       strip_spaces(source);
       if (!source[0])
           exit(fprintf(stderr,"%s:%d: empty source file name, aborting.\n",
                          fn,line));
       continue;
    } else if (!strcmp(cmd,"@header")) {
       header= strdup(cmd+strlen(cmd)+1);
       strip_spaces(header);
       if (!header[0])
           exit(fprintf(stderr,"%s:%d: empty header file name, aborting.\n",
                          fn,line));
       continue;
    } else if (!strcmp(cmd,"@pfx")) {
       pfx= strdup(cmd+strlen(cmd)+1);
       strip_spaces(pfx);
       if (!pfx[0])
           exit(fprintf(stderr,"%s:%d: empty prefix, aborting.\n", fn,line));
       continue;
    } else {
       type_t *T;
       int N= 0;
       for(T=types;T;T=T->next)
         if (!strcmp(cmd+1, T->name)) break;
       if (!T)
         exit(fprintf(stderr,"%s:%d: type '%s' not found.\n", fn, line, cmd+1));
       if (!arg_tail)
         exit(fprintf(stderr,"%s:%d: argument declaration doesn't follow "
                    "message declaration.\n", fn, line));
       begin= cmd+strlen(cmd)+1;
       while(1)
       {
         char *name;
         name= get_word(begin,end);
         if (!name) break;
         N++;
         begin= name+strlen(name)+1;
         strip_spaces(name);
         arg_t *A= malloc(sizeof(*A));
         A->type= T; A->name= strdup(name);
         A->next= 0; *arg_tail= A; arg_tail= &A->next;
       }
       if (!N)
        exit(fprintf(stderr,"%s:%d: at least one argument name must "
                 " be given in argument declaration.\n", fn, line));
       continue;
    }
  }

  fclose(f);
}


void upper(FILE *f, char *k)
{
  while(*k) fprintf(f,"%c", toupper(*(k++)));
}

void print_enum_name(FILE *f,msg_t *M)
{
  upper(f, pfx); fprintf(f,"_"); upper(f, M->name); 
}

void print_enums(FILE *f)
{
  msg_t *M;
  fprintf(f,"enum {\n");
  for(M=msgs;M;M=M->next) {
    fprintf(f,"  ");
    print_enum_name(f,M);
    fprintf(f,",\n");
  }
  fprintf(f,"};\n");
}


void print_struct_name(FILE *f, msg_t *M)
{
  fprintf(f,"%s_", pfx); upper(f,M->name); fprintf(f,"_t");
}

void print_arg(FILE *f, arg_t *a)
{
  char *k;
  type_t *T;
  T= a->type;
  k= strchr(T->cname,'$');
  if (k) {
     fwrite(T->cname, 1, k-T->cname, f);
     fprintf(f, "%s%s", a->name,k+1);
  } else fprintf(f,"%s %s",T->cname, a->name);
}

void print_typedef(FILE *f, msg_t *M)
{
  arg_t *a;
  fprintf(f,"typedef struct {\n  int kind;\n");
  for(a=M->args;a;a=a->next) {
     fprintf(f,"  ");
     print_arg(f,a);
     fprintf(f,";\n");
  }
  fprintf(f,"} ");
  print_struct_name(f,M);
  fprintf(f,";\n");
}

void print_typedefs(FILE *f)
{
  msg_t *M;
  for(M=msgs;M;M=M->next) print_typedef(f,M);
  fprintf(f, "typedef struct { int kind; } %s_t; \n", pfx);
}

void print_proto(FILE *f, msg_t *M)
{
  fprintf(f,"%s_t* %s_",pfx,pfx);
  upper(f, M->name); fprintf(f,"(");
  arg_t *a;
  int i;
  for(a=M->args,i=0;a;a=a->next,i++)
  {
     if (i) fprintf(f,",");
     print_arg(f,a);
  }
  fprintf(f,")");
}

void print_protos(FILE *f)
{
  msg_t *M;
  fprintf(f,"#ifdef __cplusplus\n");
  fprintf(f,"extern \"C\" {\n");
  fprintf(f,"#endif\n");
  for(M=msgs;M;M=M->next) {
     print_proto(f,M);
     fprintf(f,";\n");
  }
  fprintf(f,"%s_t *%s_dispatch(%s_t *);\n", pfx,pfx,pfx);
  fprintf(f,"#ifdef __cplusplus\n");
  fprintf(f,"}\n");
  fprintf(f,"#endif\n");
}

void print_body(FILE *f, msg_t *M)
{
  fprintf(f,"  "); print_struct_name(f, M);
     fprintf(f," *_YY= (typeof(_YY)) malloc(sizeof(*_YY));\n");
  fprintf(f,"  _YY->kind="); print_enum_name(f,M); fprintf(f,";\n");

  arg_t *A;
  for(A=M->args;A;A=A->next)
    fprintf(f,"  _YY->%s= %s;\n", A->name, A->name);
}

void print_bodies(FILE *f)
{
  msg_t *M;
  for(M=msgs;M;M=M->next) {
     print_proto(f,M);
     fprintf(f,"\n{\n");
     print_body(f,M);
     fprintf(f,"  return (%s_t*) _YY;\n",pfx);
     fprintf(f,"}\n");
  }
}

void print_destroy(FILE *f, char *name, char *destroy)
{
   char *k;
   if (!strcmp(destroy,".")) return ;
   fprintf(f,"    ");
   k= strchr(destroy,'$');
   if (k)
   {
     fwrite(destroy,1,k-destroy,f);
     fprintf(f,"_YY->%s%s", name, k+1);
   } else {
     fprintf(f,"%s _YY->%s", destroy, name);
   }
   fprintf(f,"\n");
}

void print_dispatch(FILE *f, msg_t *M)
{
   int i; arg_t *A;

   fprintf(f,"  case "); print_enum_name(f,M); fprintf(f,":\n");
   fprintf(f,"    {"); print_struct_name(f,M);
                                   fprintf(f," *_YY= (typeof(_YY)) _MY;\n");
   fprintf(f,"    %s_%s(", pfx,M->name);
   for(i=0,A=M->args;A;A=A->next,i++) 
    { if (i) fprintf(f,","); fprintf(f,"_YY->%s",A->name); }
   fprintf(f,");\n");
   for(A=M->args;A;A=A->next) 
     print_destroy(f,A->name, A->type->destroy);
   fprintf(f,"    free(_YY); return 0; }\n");
}

void print_dispatch_func(FILE *f)
{
   fprintf(f,"%s_t* %s_dispatch(%s_t *_MY) {\n"
             "  switch(_MY->kind) {\n", pfx,pfx,pfx);
   
   msg_t *M;
   for(M=msgs;M;M=M->next) print_dispatch(f,M);
   fprintf(f,"  }\n  return _MY;\n}\n");
}

int main(int argc,char **argv)
{
  int i;
  new_type("int", "int $", ".");
  new_type("float", "float $", ".");
  new_type("double", "double $", ".");
  new_type("str", "char* $", "free($);");
  new_type("fstr", "char *$", ".");
  
  if (argc<2) 
    exit(fprintf(stderr,"usage: %s file file .. \n",argv[0]));
  for(i=1;i<argc;i++)
   parse_file(argv[i]);


  FILE *f= fopen(header,"w");
  if (!f) exit(fprintf(stderr,"can't open %s: %s\n", header, strerror(errno)));

  print_enums(f); print_typedefs(f);
  print_protos(f);
  fclose(f);

  f= fopen(source,"w");
  if (!f) exit(fprintf(stderr,"can't open %s: %s\n", source, strerror(errno)));
  print_bodies(f);
  print_dispatch_func(f);

  fclose(f);
  return 0;
}
